""" Wrapper for clix-meta yaml file.
    This read the yaml and make its content accessible an instance of ClixMetaIndices.
    It also exposes some type hints of yaml content.
"""
import os
from pathlib import Path
from typing import Any, Dict, Optional, TypedDict

import yaml

CLIX_YAML_PATH = (
    Path(os.path.dirname(os.path.abspath(__file__))) / "index_definitions.yml"
)


class EtMetadata(TypedDict):
    short_name: str
    long_name: str
    definition: str
    comment: str


class OutputMetadata(TypedDict):
    var_name: str
    standard_name: str
    long_name: str
    units: str
    cell_methods: Dict


class ClixMetaIndex(TypedDict):
    reference: str
    period: Dict
    output: OutputMetadata
    input: Dict
    index_function: Dict
    ET: EtMetadata


class ClixMetaIndices:
    """
    Singleton to access content of clix-meta yaml file.
    """

    __instance: Any = None
    indices_record: Dict[str, ClixMetaIndex]

    @staticmethod
    def get_instance():
        if ClixMetaIndices.__instance is None:
            ClixMetaIndices.__instance = ClixMetaIndices()
        return ClixMetaIndices.__instance

    def __init__(self):
        if ClixMetaIndices.__instance is not None:
            raise Exception("This class is a singleton! Use Clix::get_instance.")
        else:
            ClixMetaIndices.__instance = self
            with open(CLIX_YAML_PATH, "r") as clix_meta_file:
                self.indices_record = yaml.safe_load(clix_meta_file)["indices"]

    def lookup(self, query: str) -> Optional[ClixMetaIndex]:
        for index in self.indices_record.keys():
            if index.upper() == query.upper():
                return self.indices_record[index]
        return None



"""
    `icclim.models.frequency` wraps the concept of pandas frequency in order to resample
    time series.  `slice_mode` paramater of `icclim.index` is always converted to a
    `Frequency`.
"""
from __future__ import annotations

import datetime
from enum import Enum
from typing import Callable, List, Tuple, Union

import cftime
import numpy as np
import pandas as pd
import xarray as xr
from xarray.core.dataarray import DataArray

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import (
    AMJJAS_MONTHS,
    DJF_MONTHS,
    JJA_MONTHS,
    MAM_MONTHS,
    MONTHS_MAP,
    ONDJFM_MONTHS,
    SON_MONTHS,
)


def get_month_filter(month_list: list[int]) -> Callable:
    return lambda da: filter_months(da, month_list)


def filter_months(da: DataArray, month_list: list[int]) -> DataArray:
    return da.sel(time=da.time.dt.month.isin(month_list))


def get_seasonal_time_updater(
    start_month: int,
    end_month: int,
) -> Callable[[DataArray], tuple[DataArray, DataArray]]:
    """Seasonal time updater and time bounds creator method generator.
    Returns a callable of DataArray which will rewrite the time dimension to
    the season composed of the given month. The data must have been computed on this
    season beforehand.
    It also create the corresponding time_bounds.
    Parameters
    ----------
    start_month: int
        The season starting month, it must be between 1 and 12.
    end_month: int
        The season ending month, it must be between 1 and 12.
    Returns
    -------
    function: Callable[[DataArray], DataArray]
        function resampling the input da to the wanted season.
    """

    def add_time_bounds(da: DataArray) -> tuple[DataArray, DataArray]:
        da_years = np.unique(da.time.dt.year)
        time_bounds = []
        new_time_axis = []
        first_time = da.time.values[0]
        for year in da_years:
            if start_month > end_month:
                year_of_season_end = year + 1
            else:
                year_of_season_end = year
            if isinstance(first_time, cftime.datetime):
                start = cftime.datetime(
                    year, start_month, 1, calendar=first_time.calendar
                )
                end = cftime.datetime(
                    year_of_season_end, end_month + 1, 1, calendar=first_time.calendar
                )
            else:
                start = pd.to_datetime(f"{year}-{start_month}")
                end = pd.to_datetime(f"{year_of_season_end}-{end_month + 1}")
            end = end - datetime.timedelta(days=1)
            new_time_axis.append(start + (end - start) / 2)
            time_bounds.append([start, end])
        da.coords["time"] = ("time", new_time_axis)
        time_bounds_da = DataArray(
            data=time_bounds,
            dims=["time", "bounds"],
            coords=[("time", da.time.values), ("bounds", [0, 1])],
        )
        return da, time_bounds_da

    return add_time_bounds


def _get_time_bounds_updater(
    freq: str,
) -> Callable[[DataArray], tuple[DataArray, DataArray]]:
    def add_time_bounds(da: DataArray) -> tuple[DataArray, DataArray]:
        # da should already be resampled to freq
        if isinstance(da.indexes.get("time"), xr.CFTimeIndex):
            offset = xr.coding.cftime_offsets.to_offset(freq)
            starts = np.array(
                [
                    cftime.datetime(
                        date.year,
                        date.month,
                        date.day,
                        date.hour,
                        date.minute,
                        date.second,
                        calendar=date.calendar,
                    )
                    for date in da.indexes.get("time")
                ]
            )
            ends = starts + offset
            ends = ends - datetime.timedelta(days=1)
        else:
            offset = pd.tseries.frequencies.to_offset(freq)
            starts = pd.to_datetime(da.time.dt.floor("D"))
            ends = starts + offset
            ends = ends - pd.Timedelta(days=1)
        # make time axis values be in the middle of the bounds
        da["time"] = starts + (ends - starts) / 2
        time_bounds_da = DataArray(
            data=list(zip(starts, ends)),
            dims=["time", "bounds"],
            coords=[("time", da.time.values), ("bounds", [0, 1])],
        )
        return da, time_bounds_da

    return add_time_bounds


class _Freq:
    """Internal class to ease writing and maintaining the enum.
    Without it, in the instanciation of enum values we would have to write tuples
    would not be able to use kwargs, which make the code less readable.
    """

    def __init__(
        self,
        panda_freq: str,
        accepted_values: list[str],
        description: str,
        post_processing: Callable[[DataArray], tuple[DataArray, DataArray]],
        pre_processing: Callable[[DataArray], DataArray],
    ):
        self.panda_freq: str = panda_freq
        self.accepted_values: list[str] = accepted_values
        self.description = description
        self.post_processing = post_processing
        self.pre_processing = pre_processing


class Frequency(Enum):
    """The sampling frequency of the resulting dataset."""

    MONTH = _Freq(
        panda_freq="MS",
        accepted_values=["month", "MS"],
        description="monthly time series",
        post_processing=_get_time_bounds_updater("MS"),
        pre_processing=lambda x: x,
    )
    """ Resample to monthly values"""

    AMJJAS = _Freq(
        panda_freq="AS-APR",
        accepted_values=["AMJJAS"],
        description="summer half-year time series",
        post_processing=get_seasonal_time_updater(AMJJAS_MONTHS[0], AMJJAS_MONTHS[-1]),
        pre_processing=get_month_filter(AMJJAS_MONTHS),
    )
    """ Resample to summer half-year, from April to September included."""

    ONDJFM = _Freq(
        panda_freq="AS-OCT",
        accepted_values=["ONDJFM"],
        description="winter half-year time series",
        post_processing=get_seasonal_time_updater(ONDJFM_MONTHS[0], ONDJFM_MONTHS[-1]),
        pre_processing=get_month_filter(ONDJFM_MONTHS),
    )
    """ Resample to winter half-year, from October to March included."""

    DJF = _Freq(
        panda_freq="AS-DEC",
        accepted_values=["DJF"],
        description="winter time series",
        post_processing=get_seasonal_time_updater(DJF_MONTHS[0], DJF_MONTHS[-1]),
        pre_processing=get_month_filter(DJF_MONTHS),
    )
    """ Resample to winter season, from December to February included."""

    MAM = _Freq(
        panda_freq="AS-MAR",
        accepted_values=["MAM"],
        description="spring time series",
        post_processing=get_seasonal_time_updater(MAM_MONTHS[0], MAM_MONTHS[-1]),
        pre_processing=get_month_filter(MAM_MONTHS),
    )
    """ Resample to spring season, from March to May included."""

    JJA = _Freq(
        panda_freq="AS-JUN",
        accepted_values=["JJA"],
        description="summer time series",
        post_processing=get_seasonal_time_updater(JJA_MONTHS[0], JJA_MONTHS[-1]),
        pre_processing=get_month_filter(JJA_MONTHS),
    )
    """ Resample to summer season, from June to Agust included."""

    SON = _Freq(
        panda_freq="AS-SEP",
        accepted_values=["SON"],
        description="autumn time series",
        post_processing=get_seasonal_time_updater(SON_MONTHS[0], SON_MONTHS[-1]),
        pre_processing=get_month_filter(SON_MONTHS),
    )
    """ Resample to fall season, from September to November included."""

    CUSTOM = _Freq(
        panda_freq="MS",
        accepted_values=[],
        description="",
        post_processing=lambda x: x,
        pre_processing=lambda x: x,
    )
    """ Placeholder instance for custom sampling frequencies.
        Do not use as is, use `slice_mode` with "month", "season" or "dates" keywords
        instead.
    """

    YEAR = _Freq(
        panda_freq="YS",
        accepted_values=["year", "YS"],
        description="annual time series",
        post_processing=_get_time_bounds_updater("YS"),
        pre_processing=lambda x: x,
    )
    """ Resample to yearly values."""

    def __init__(self, freq: _Freq):
        self._freq = freq

    @property
    def panda_freq(self):
        return self._freq.panda_freq

    @property
    def accepted_values(self):
        return self._freq.accepted_values

    @property
    def description(self):
        return self._freq.description

    @property
    def post_processing(self):
        return self._freq.post_processing

    @property
    def pre_processing(self):
        return self._freq.pre_processing

    @staticmethod
    def lookup(slice_mode: SliceMode) -> Frequency:
        if isinstance(slice_mode, Frequency):
            return slice_mode
        if isinstance(slice_mode, str):
            return _get_frequency_from_string(slice_mode)
        if isinstance(slice_mode, list):
            return _get_frequency_from_list(slice_mode)
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {slice_mode}."
            f"Use a Frequency from {[f for f in Frequency]}"
        )

    @staticmethod
    def is_seasonal(slice_mode: SliceMode) -> bool:
        return Frequency.lookup(slice_mode) in [
            Frequency.CUSTOM,
            Frequency.ONDJFM,
            Frequency.AMJJAS,
            Frequency.MAM,
            Frequency.JJA,
            Frequency.SON,
            Frequency.DJF,
        ]


def _get_frequency_from_string(slice_mode: str) -> Frequency:
    for freq in Frequency:
        if freq.name == slice_mode.upper() or slice_mode.upper() in map(
            str.upper, freq.accepted_values
        ):
            return freq
    # TODO: we could add a compatibility to other pandas freq if we detect
    #       something like WS, 4MS, etc. In which case we would use FREQUENCY.CUSTOM
    raise InvalidIcclimArgumentError(f"Unknown frequency {slice_mode}.")


def _is_season_valid(months):
    is_valid = True
    for i in range(0, len(months) - 1):
        is_valid = is_valid and months[i] > 0 and months[i] < 13
        if months[i] > months[i + 1]:
            is_valid = is_valid and months[i + 1] == 1 and months[i] == 12
        else:
            is_valid = is_valid and (months[i + 1] - months[i] == 1)
    return is_valid


def _get_frequency_from_list(slice_mode_list: list) -> Frequency:
    if len(slice_mode_list) < 2:
        raise InvalidIcclimArgumentError(
            f"The given slice list {slice_mode_list}"
            f" has a length of {len(slice_mode_list)}."
            f" The maximum length here is 2."
        )
    sampling_freq = slice_mode_list[0]
    custom_freq = Frequency.CUSTOM
    if sampling_freq in ["month", "months"]:
        months = slice_mode_list[1]

        def month_list_post_processing(da):
            res, bounds = _get_time_bounds_updater("MS")(da)
            res = get_month_filter(months)(res)
            return res, bounds

        custom_freq._freq = _Freq(
            pre_processing=get_month_filter(months),
            post_processing=month_list_post_processing,
            panda_freq="MS",
            description=f"monthly time series (months: {months})",
            accepted_values=[],
        )
    elif sampling_freq == "season":
        months = slice_mode_list[1]
        if isinstance(months, Tuple):
            months = months[0] + months[1]  # concat in case of ([12], [1, 2])
        if not _is_season_valid(months):
            raise InvalidIcclimArgumentError(
                f"A season created using `slice_mode` must be made of consecutive"
                f" months. It was {months}."
            )
        custom_freq._freq = _Freq(
            pre_processing=get_month_filter(months),
            post_processing=get_seasonal_time_updater(months[0], months[-1]),
            panda_freq=f"AS-{MONTHS_MAP[months[0]]}",
            description=f"seasonal time series (season: {months})",
            accepted_values=[],
        )
    else:
        raise InvalidIcclimArgumentError(
            f"Unknown frequency {slice_mode_list}."
            " The sampling frequency must be one of {'season', 'month'}"
        )
    return custom_freq


SliceMode = Union[Frequency, str, List[Union[str, Tuple, int]]]



import os
import shutil

import numpy as np
import pandas as pd
import pytest
import xarray as xr

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.ecad_indices import EcadIndex
from icclim.pre_processing.input_parsing import read_dataset, update_to_standard_coords


def test_update_to_standard_coords():
    # GIVEN
    ds = xr.Dataset(
        {
            "pouet": xr.DataArray(
                data=np.full(10, 42).reshape((10, 1, 1)),
                coords=dict(
                    latitude=[42],
                    longitude=[42],
                    t=pd.date_range("2042-01-01", periods=10, freq="D"),
                ),
                dims=["t", "latitude", "longitude"],
                name="pr",
                attrs={"units": "kg m-2 d-1"},
            )
        }
    )
    # WHEN
    res, revert = update_to_standard_coords(ds)
    # THEN
    assert "lat" in res.coords
    assert "time" in res.coords
    assert "lon" in res.coords
    assert res.rename(revert).coords.keys() == ds.coords.keys()


class Test_ReadDataset:
    OUTPUT_NC_FILE = "tmp.nc"
    OUTPUT_NC_FILE_2 = "tmp-2.nc"
    OUTPUT_ZARR_STORE = "tmp.zarr"
    OUTPUT_UNKNOWN_FORMAT = "tmp.cacahuete"

    @pytest.fixture(autouse=True)
    def cleanup(self):
        # setup
        yield
        # teardown
        shutil.rmtree(self.OUTPUT_ZARR_STORE, ignore_errors=True)
        for f in [
            self.OUTPUT_NC_FILE,
            self.OUTPUT_NC_FILE_2,
            self.OUTPUT_UNKNOWN_FORMAT,
        ]:
            try:
                os.remove(f)
            except FileNotFoundError:
                pass

    def test_read_dataset_xr_da_user_index_error(self):
        da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                t=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["t", "latitude", "longitude"],
            name="pr",
            attrs={"units": "kg m-2 d-1"},
        )
        with pytest.raises(InvalidIcclimArgumentError):
            read_dataset(da)

    def test_read_dataset_xr_da_ecad_index_error(self):
        da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                t=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["t", "latitude", "longitude"],
            name="pr",
            attrs={"units": "kg m-2 d-1"},
        )
        with pytest.raises(InvalidIcclimArgumentError):
            read_dataset(da, EcadIndex.WW)

    def test_read_dataset_xr_da_ecad_index_success(self):
        da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                t=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["t", "latitude", "longitude"],
            name="pr",
            attrs={"units": "kg m-2 d-1"},
        )
        ds_res, chunk_it, is_zarr = read_dataset(da, EcadIndex.TX90P)
        xr.testing.assert_equal(ds_res.tasmax, da)
        assert chunk_it is False
        assert is_zarr is False

    def test_read_dataset_xr_da_user_index_success(self):
        da = xr.DataArray(
            data=np.full(10, 42).reshape((10, 1, 1)),
            coords=dict(
                latitude=[42],
                longitude=[42],
                t=pd.date_range("2042-01-01", periods=10, freq="D"),
            ),
            dims=["t", "latitude", "longitude"],
            name="pr",
            attrs={"units": "kg m-2 d-1"},
        )
        ds_res, chunk_it, is_zarr = read_dataset(da, None, "doto")
        xr.testing.assert_equal(ds_res.doto, da)
        assert chunk_it is False
        assert is_zarr is False

    def test_read_dataset_xr_ds_success(self):
        ds = xr.Dataset(
            {
                "pouet": xr.DataArray(
                    data=np.full(10, 42).reshape((10, 1, 1)),
                    coords=dict(
                        latitude=[42],
                        longitude=[42],
                        t=pd.date_range("2042-01-01", periods=10, freq="D"),
                    ),
                    dims=["t", "latitude", "longitude"],
                    name="pr",
                    attrs={"units": "kg m-2 d-1"},
                )
            }
        )
        ds_res, chunk_it, is_zarr = read_dataset(ds)
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)
        assert chunk_it is False
        assert is_zarr is False

    def test_read_dataset_netcdf_success(self):
        ds = xr.Dataset(
            {
                "pouet": xr.DataArray(
                    data=np.full(10, 42).reshape((10, 1, 1)),
                    coords=dict(
                        latitude=[42],
                        longitude=[42],
                        t=pd.date_range("2042-01-01", periods=10, freq="D"),
                    ),
                    dims=["t", "latitude", "longitude"],
                    name="pr",
                    attrs={"units": "kg m-2 d-1"},
                )
            }
        )
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        ds_res, chunk_it, is_zarr = read_dataset(self.OUTPUT_NC_FILE)
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)
        assert chunk_it is True
        assert is_zarr is False

    def test_read_dataset_multi_netcdf_success(self):
        ds = xr.Dataset(
            {
                "pouet": xr.DataArray(
                    data=np.full(10, 42).reshape((10, 1, 1)),
                    coords=dict(
                        latitude=[42],
                        longitude=[42],
                        t=pd.date_range("2042-01-01", periods=10, freq="D"),
                    ),
                    dims=["t", "latitude", "longitude"],
                    name="pr",
                    attrs={"units": "kg m-2 d-1"},
                )
            }
        )
        ds.to_netcdf(self.OUTPUT_NC_FILE)
        ds.rename({"pouet": "patapouet"}).to_netcdf(self.OUTPUT_NC_FILE_2)
        # WHEN
        ds_res, chunk_it, is_zarr = read_dataset(
            [self.OUTPUT_NC_FILE, self.OUTPUT_NC_FILE_2]
        )
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)
        xr.testing.assert_equal(ds_res.patapouet, ds.pouet)
        assert chunk_it is True
        assert is_zarr is False

    def test_read_dataset_zarr_store_success(self):
        ds = xr.Dataset(
            {
                "pouet": xr.DataArray(
                    data=np.full(10, 42).reshape((10, 1, 1)),
                    coords=dict(
                        latitude=[42],
                        longitude=[42],
                        t=pd.date_range("2042-01-01", periods=10, freq="D"),
                    ),
                    dims=["t", "latitude", "longitude"],
                    name="pr",
                    attrs={"units": "kg m-2 d-1"},
                )
            }
        )
        ds.to_zarr(self.OUTPUT_ZARR_STORE)
        # WHEN
        ds_res, chunk_it, is_zarr = read_dataset(self.OUTPUT_ZARR_STORE)
        # THEN
        xr.testing.assert_equal(ds_res.pouet, ds.pouet)
        assert chunk_it is True
        assert is_zarr is True

    def test_read_dataset_not_implemented_error(self):
        # WHEN
        with pytest.raises(NotImplementedError):
            read_dataset(42)  # noqa



from xarray import DataArray, Dataset


# FIXME To remove once minimal xarray version is v0.20.0 (use .chunksizes instead)
def _da_chunksizes(da: DataArray) -> dict:
    # Copied and adapted from xarray
    if hasattr(da.data, "chunks"):
        return {dim: c for dim, c in zip(da.dims, da.data.chunks)}
    else:
        return {}


# FIXME To remove once minimal xarray version is v0.20.0 (use .chunksizes instead)
def _get_chunksizes(ds: Dataset) -> dict:
    # Copied and adapted from xarray
    chunks = {}
    for v in ds.variables.values():
        if hasattr(v.data, "chunks"):
            for dim, c in _da_chunksizes(v).items():
                if dim in chunks and c != chunks[dim]:
                    raise ValueError(
                        f"Object has inconsistent chunks along dimension {dim}."
                        " This can be fixed by calling unify_chunks()."
                    )
                chunks[dim] = c
    return chunks



from typing import Callable
from unittest.mock import MagicMock, patch

import pytest

from icclim.icclim_exceptions import InvalidIcclimArgumentError, MissingIcclimInputError
from icclim.models.constants import PRECIPITATION, TEMPERATURE
from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable
from icclim.models.user_index_config import LogicalOperation
from icclim.tests.test_utils import stub_pr, stub_tas, stub_user_index
from icclim.user_indices import calc_operation
from icclim.user_indices.calc_operation import (
    CalcOperation,
    anomaly,
    compute_user_index,
    count_events,
    max_consecutive_event_count,
    run_mean,
    run_sum,
)


class Test_compute:
    def test_error_bad_operation(self):
        # GIVEN
        cf_var = CfVariable("tas", stub_tas(), stub_tas())
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "pouet pouet"
        user_index.freq = Frequency.MONTH
        # WHEN
        with pytest.raises(InvalidIcclimArgumentError):
            compute_user_index(user_index)

    def test_simple(self):
        # GIVEN
        cf_var = CfVariable("tas", stub_tas(), stub_tas())
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "max"
        user_index.freq = Frequency.MONTH
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 1

    def test_simple_percentile_pr(self):
        # GIVEN
        cf_var = CfVariable("tas", stub_pr(5), stub_pr(5))
        cf_var.study_da.data[15:30] += 10
        cf_var.study_da.data[366 + 15 : 366 + 30] = 2  # Ignore because not in base
        cf_var.reference_da = cf_var.study_da.sel(
            time=cf_var.study_da.time.dt.year == 2042
        )
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = CalcOperation.MIN
        user_index.thresh = "90p"
        user_index.logical_operation = LogicalOperation.GREATER_OR_EQUAL_THAN
        user_index.var_type = PRECIPITATION
        user_index.freq = Frequency.YEAR
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 5

    def test_simple_percentile_temp(self):
        cf_var = CfVariable("tas", stub_tas(5), stub_tas(5))
        cf_var.study_da.data[15:30] = 1
        cf_var.reference_da = cf_var.study_da.sel(
            time=cf_var.study_da.time.dt.year.isin([2042, 2043])
        )
        user_index = stub_user_index([cf_var])
        user_index.calc_operation = "min"
        user_index.thresh = "10p"
        user_index.logical_operation = LogicalOperation.LOWER_OR_EQUAL_THAN
        user_index.var_type = TEMPERATURE
        user_index.freq = Frequency.MONTH
        # WHEN
        result = compute_user_index(user_index)
        # THEN
        assert result.data[0] == 1
        assert result.data[1] == 5

    @patch("icclim.models.user_index_config.UserIndexConfig")
    @patch("icclim.models.index_config.CfVariable")
    def test_error_anomaly(self, config_mock: MagicMock, cf_var_mock: MagicMock):
        config_mock.cf_vars = [cf_var_mock]
        cf_var_mock.reference_da = None
        with pytest.raises(MissingIcclimInputError):
            anomaly(config_mock)

    @patch("icclim.models.user_index_config.UserIndexConfig")
    @patch("icclim.user_indices.operators.anomaly")
    @patch("icclim.models.index_config.CfVariable")
    def test_success_anomaly(
        self, config_mock: MagicMock, op_mock: MagicMock, cf_var_mock: MagicMock
    ):
        config_mock.cf_vars = [cf_var_mock]
        cf_var_mock.reference_da = [1, 2, 3]  # no-op, just need to mock a valid length
        anomaly(config_mock)
        op_mock.assert_called_once()

    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_error_run_sum(self, config_mock: MagicMock):
        config_mock.extreme_mode = None
        with pytest.raises(MissingIcclimInputError):
            run_sum(config_mock)
        config_mock.extreme_mode = {}
        config_mock.window_width = None
        with pytest.raises(MissingIcclimInputError):
            run_sum(config_mock)

    @patch("icclim.user_indices.operators.run_sum")
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_success_run_sum(self, config_mock: MagicMock, op_mock: MagicMock):
        run_sum(config_mock)
        op_mock.assert_called_once()

    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_error_run_mean(self, config_mock: MagicMock):
        config_mock.extreme_mode = None
        with pytest.raises(MissingIcclimInputError):
            run_mean(config_mock)
        config_mock.extreme_mode = {}
        config_mock.window_width = None
        with pytest.raises(MissingIcclimInputError):
            run_mean(config_mock)

    @patch("icclim.user_indices.operators.run_mean")
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_success_run_mean(self, config_mock: MagicMock, op_mock: MagicMock):
        run_mean(config_mock)
        op_mock.assert_called_once()

    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_error_max_consecutive_event_count(self, config_mock: MagicMock):
        config_mock.logical_operation = None
        with pytest.raises(MissingIcclimInputError):
            max_consecutive_event_count(config_mock)
        config_mock.logical_operation = {}
        config_mock.thresh = None
        with pytest.raises(MissingIcclimInputError):
            max_consecutive_event_count(config_mock)
        config_mock.logical_operation = {}
        config_mock.thresh = []
        with pytest.raises(InvalidIcclimArgumentError):
            max_consecutive_event_count(config_mock)

    @patch("icclim.user_indices.operators.max_consecutive_event_count")
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_success_max_consecutive_event_count(
        self, config_mock: MagicMock, op_mock: MagicMock
    ):
        max_consecutive_event_count(config_mock)
        op_mock.assert_called_once()

    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_error_count_events(self, config_mock: MagicMock):
        config_mock.nb_event_config = None
        with pytest.raises(MissingIcclimInputError):
            count_events(config_mock)

    @patch("icclim.user_indices.operators.count_events")
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_success_count_events(self, config_mock: MagicMock, op_mock: MagicMock):
        count_events(config_mock)
        op_mock.assert_called_once()

    @pytest.mark.parametrize(
        "reducer",
        [
            calc_operation.sum,
            calc_operation.mean,
            calc_operation.min,
            calc_operation.max,
        ],
    )
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_error_simple_reducer(self, config_mock: MagicMock, reducer: Callable):
        config_mock.cf_vars = [1, 2, 3]
        with pytest.raises(InvalidIcclimArgumentError):
            reducer(config_mock)
        config_mock.cf_vars = [MagicMock()]
        config_mock.thresh = []
        with pytest.raises(InvalidIcclimArgumentError):
            reducer(config_mock)

    @pytest.mark.parametrize("reducer", ["sum", "mean", "min", "max"])
    @patch("icclim.models.user_index_config.UserIndexConfig")
    def test_success_simple_reducer(self, config_mock: MagicMock, reducer: str):
        config_mock.calc_operation = reducer
        config_mock.cf_vars = [MagicMock()]
        config_mock.thresh = 42
        with patch("icclim.user_indices.operators." + reducer) as op_mock:
            compute_user_index(config_mock)
            op_mock.assert_called_once()



import pytest

from icclim.models.index_group import IndexGroup


@pytest.mark.parametrize(
    "gr",
    [
        ("temperature", IndexGroup.TEMPERATURE),
        ("heat", IndexGroup.HEAT),
        ("cold", IndexGroup.COLD),
        ("drought", IndexGroup.DROUGHT),
        ("rain", IndexGroup.RAIN),
        ("snow", IndexGroup.SNOW),
        ("compound", IndexGroup.COMPOUND),
    ],
)
def test_lookup_success(gr):
    assert IndexGroup.lookup(gr[0]) is gr[1]


def test_lookup_error():
    with pytest.raises(NotImplementedError):
        IndexGroup.lookup("coin coin le canard")



# -*- Coding: latin-1 -*-
#  Copyright CERFACS (http://cerfacs.fr/)
#  Apache License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
"""
Main module of icclim.
"""
from __future__ import annotations

import logging
import time
from datetime import datetime
from typing import Callable, Literal
from warnings import warn

import xarray as xr
import xclim
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

from icclim.ecad_functions import IndexConfig
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_logger import IcclimLogger, Verbosity
from icclim.models.ecad_indices import EcadIndex
from icclim.models.frequency import Frequency, SliceMode
from icclim.models.index_group import IndexGroup
from icclim.models.netcdf_version import NetcdfVersion
from icclim.models.quantile_interpolation import QuantileInterpolation
from icclim.models.user_index_config import UserIndexConfig
from icclim.models.user_index_dict import UserIndexDict
from icclim.pre_processing.input_parsing import read_dataset, update_to_standard_coords
from icclim.user_indices.calc_operation import CalcOperation, compute_user_index

log: IcclimLogger = IcclimLogger.get_instance(Verbosity.LOW)


def indices(
    index_group: Literal["all"] | str | IndexGroup | list[str],
    ignore_error: bool = False,
    **kwargs,
) -> Dataset:
    """
    Compute multiple indices at the same time.
    The input dataset(s) must include all the necessary variables.
    It can only be used with keyword arguments (kwargs)

    Parameters
    ----------
    index_group : "all" | str | IndexGroup | list[str]
        Either the name of an IndexGroup, a instance of IndexGroup or a list
        of index short names.
        The value "all" can also be used to compute every indices.
        Note that the input given by ``in_files`` must include all the necessary
        variables to compute the indices of this group.
    kwargs : Dict
        ``icclim.index`` keyword arguments.

    Returns
    -------
    xr.Dataset
        A Dataset with one data variable per index.

    .. notes
        If ``output_file`` is part of kwargs, the result is written in a single netCDF
        file, which will contain all the index results of this group.

    """
    if isinstance(index_group, list):
        indices = [EcadIndex.lookup(i) for i in index_group]
    elif index_group == IndexGroup.WILD_CARD_GROUP or (
        isinstance(index_group, str)
        and index_group.lower() == IndexGroup.WILD_CARD_GROUP.value
    ):
        indices = iter(EcadIndex)
    else:
        indices = IndexGroup.lookup(index_group).get_indices()
    out = None
    if "out_file" in kwargs.keys():
        out = kwargs["out_file"]
        del kwargs["out_file"]
    acc = []
    for i in indices:
        kwargs["index_name"] = i.short_name
        if ignore_error:
            try:
                acc.append(index(**kwargs))
            except Exception:
                warn(f"Could not compute {i.short_name}.")
        else:
            acc.append(index(**kwargs))
    ds: Dataset = xr.merge(acc)
    if out is not None:
        _write_output_file(
            result_ds=ds,
            input_time_encoding=ds.time.encoding,
            netcdf_version=kwargs.get("netcdf_version", NetcdfVersion.NETCDF4),
            file_path=out,
        )
    return ds


def indice(*args, **kwargs):
    """
    Deprecated proxy for `icclim.index` function.
    To be deleted in a futur version.
    """
    log.deprecation_warning(old="icclim.indice", new="icclim.index")
    return index(*args, **kwargs)


def index(
    in_files: str | list[str] | Dataset | DataArray,
    index_name: str | None = None,  # optional when computing user_indices
    var_name: str | list[str] | None = None,
    slice_mode: SliceMode = Frequency.YEAR,
    time_range: list[datetime] = None,
    out_file: str | None = None,
    threshold: float | list[float] | None = None,
    callback: Callable[[int], None] = log.callback,
    callback_percentage_start_value: int = 0,
    callback_percentage_total: int = 100,
    base_period_time_range: list[datetime] | None = None,
    window_width: int = 5,
    only_leap_years: bool = False,
    ignore_Feb29th: bool = False,
    interpolation: (
        str | QuantileInterpolation | None
    ) = QuantileInterpolation.MEDIAN_UNBIASED,
    out_unit: str | None = None,
    netcdf_version: str | NetcdfVersion = NetcdfVersion.NETCDF4,
    user_index: UserIndexDict = None,
    save_percentile: bool = False,
    logs_verbosity: Verbosity | str = Verbosity.LOW,
    # deprecated parameters
    indice_name: str = None,
    user_indice: UserIndexDict = None,
    transfer_limit_Mbytes: float = None,
) -> Dataset:
    """
    Parameters
    ----------
    in_files : str | list[str] | Dataset | DataArray,
        Absolute path(s) to NetCDF dataset(s), including OPeNDAP URLs,
        or path to zarr store, or xarray.Dataset or xarray.DataArray.
    index_name : str
        Climate index name.
        For ECA&D index, case insensitive name used to lookup the index.
        For user index, it's the name of the output variable.
    var_name : str | list[str] | None
        ``optional`` Target variable name to process corresponding to ``in_files``.
        If None (default) on ECA&D index, the variable is guessed based on the climate
        index wanted.
        Mandatory for a user index.
    slice_mode : str
        Type of temporal aggregation:
        {"year", "month", "DJF", "MAM", "JJA", "SON", "ONDJFM" or "AMJJAS"}.
        Default is "year".
        See :ref:`slice_mode` for details.
    time_range : list[datetime.datetime]
        ``optional`` Temporal range: upper and lower bounds for temporal subsetting.
        If ``None``, whole period of input files will be processed.
        Default is ``None``.
    out_file : str | None
        Output NetCDF file name (default: "icclim_out.nc" in the current directory).
        Default is "icclim_out.nc".
        If the input ``in_files`` is a ``Dataset``, ``out_file`` field is ignored.
        Use the function returned value instead to retrieve the computed value.
        If ``out_file`` already exists, icclim will overwrite it!
    threshold : float | list[float] | None
        ``optional`` User defined threshold for certain indices.
        Default depend on the index, see their individual definition.
        When a list of threshold is provided, the index will be computed for each
        thresholds.
    transfer_limit_Mbytes : float
        Deprecated, does not have any effect.
    callback : Callable[[int], None]
        ``optional`` Progress bar printing. If ``None``, progress bar will not be
        printed.
    callback_percentage_start_value : int
        ``optional`` Initial value of percentage of the progress bar (default: 0).
    callback_percentage_total : int
        ``optional`` Total percentage value (default: 100).
    base_period_time_range : list[datetime.datetime]
        ``optional`` Temporal range of the reference period on which percentiles are
        computed.
        When missing, the studied period is used to compute percentiles.
        The study period is either the dataset filtered by `time_range` or the whole
        dataset if  `time_range` is None.
        On temperature based indices relying on percentiles (TX90p, WSDI...), the
        overlapping period between `base_period_time_range` and the study period is
        bootstrapped.
        On indices not relying on percentiles, this parameter is ignored.
    window_width : int
        ``optional`` User defined window width for related indices (default: 5).
        Ignored for non related indices.
    only_leap_years : bool
        ``optional`` Option for February 29th (default: False).
    ignore_Feb29th : bool
        ``optional`` Ignoring or not February 29th (default: False).
    interpolation : str | QuantileInterpolation | None
        ``optional`` Interpolation method to compute percentile values:
        ``{"linear", "hyndman_fan"}``
        Default is "hyndman_fan", a.k.a type 8 or method 8.
        Ignored for non percentile based indices.
    out_unit : str | None
        ``optional`` Output unit for certain indices: "days" or "%" (default: "days").
    netcdf_version : str | icclim.models.netcdf_version.NetcdfVersion
        ``optional`` NetCDF version to create (default: "NETCDF3_CLASSIC").
    user_index : UserIndexDict
        ``optional`` A dictionary with parameters for user defined index.
        See :ref:`Custom indices`.
        Ignored for ECA&D indices.
    save_percentile : bool
        ``optional`` True if the percentiles should be saved within the resulting netcdf
         file (default: False).
    logs_verbosity : str | Verbosity
        ``optional`` Configure how verbose icclim is.
        Possible values: ``{"LOW", "HIGH", "SILENT"}`` (default: "LOW")
    indice_name : str | None
        DEPRECATED, use index_name instead.
    user_indice : dict | None
        DEPRECATED, use user_index instead.

    """
    _setup(callback, callback_percentage_start_value, logs_verbosity, slice_mode)
    index_name, user_index = _handle_deprecated_params(
        index_name, indice_name, transfer_limit_Mbytes, user_index, user_indice
    )
    # -- Choose index to compute
    if user_index is None and index_name is None:
        raise InvalidIcclimArgumentError(
            "No index to compute."
            " You must provide either `user_index` to compute a customized index"
            " or `index_name` for one of the ECA&D indices."
        )
    index: EcadIndex | None
    if index_name is not None:
        index = EcadIndex.lookup(index_name)
    else:
        index = None
    input_dataset, chunk_it, _ = read_dataset(in_files, index, var_name)
    ds, reset_coords_dict = update_to_standard_coords(input_dataset)
    config = IndexConfig(
        base_period_time_range=base_period_time_range,
        ds=ds,
        ignore_Feb29th=ignore_Feb29th,
        only_leap_years=only_leap_years,
        save_percentile=save_percentile,
        slice_mode=slice_mode,
        time_range=time_range,
        var_names=_guess_variable_names(var_name, index, ds),
        window_width=window_width,
        out_unit=out_unit,
        netcdf_version=netcdf_version,
        interpolation=interpolation,
        callback=callback,
        index=index,
        chunk_it=chunk_it,
    )
    if user_index is not None:
        result_ds = _compute_user_index_dataset(config=config, user_index=user_index)
    else:
        result_ds = _compute_ecad_index_dataset(
            config=config,
            index=index,
            threshold=threshold,
            current_history=ds.attrs.get("history", None),
        )
    if reset_coords_dict:
        result_ds = result_ds.rename(reset_coords_dict)
    if out_file is not None:
        _write_output_file(result_ds, ds.time.encoding, config.netcdf_version, out_file)
    callback(callback_percentage_total)
    log.ending_message(time.process_time())
    return result_ds


def _write_output_file(
    result_ds: xr.Dataset,
    input_time_encoding: dict,
    netcdf_version: NetcdfVersion,
    file_path: str,
) -> None:
    """
    Write `result_ds` to a netCDF file on `out_file` path.
    """
    if input_time_encoding:
        time_encoding = {
            "calendar": input_time_encoding.get("calendar"),
            "units": input_time_encoding.get("units"),
            "dtype": input_time_encoding.get("dtype"),
        }
    else:
        time_encoding = {"units": "days since 1850-1-1"}
    result_ds.to_netcdf(
        file_path,
        format=netcdf_version.value,
        encoding={"time": time_encoding},
    )


def _handle_deprecated_params(
    index_name, indice_name, transfer_limit_Mbytes, user_index, user_indice
) -> tuple[str, UserIndexDict]:
    if indice_name is not None:
        log.deprecation_warning(old="indice_name", new="index_name")
        index_name = indice_name
    if user_indice is not None:
        log.deprecation_warning(old="user_indice", new="user_index")
        user_index = user_indice
    if transfer_limit_Mbytes is not None:
        log.deprecation_warning(old="transfer_limit_Mbytes")
    return index_name, user_index


def _setup(callback, callback_start_value, logs_verbosity, slice_mode):
    # make xclim input daily check a warning instead of an error
    # TODO: it might be safer to feed a context manager which will setup
    #       and teardown these confs
    xclim.set_options(data_validation="warn")
    if Frequency.is_seasonal(slice_mode):
        # for now seasonal slice_modes missing values cannot be checked
        xclim.set_options(check_missing="skip")
    # keep attributes through xarray operations
    xr.set_options(keep_attrs=True)
    log.set_verbosity(logs_verbosity)
    log.start_message()
    callback(callback_start_value)


def _compute_ecad_index_dataset(
    config: IndexConfig,
    index: EcadIndex,
    threshold: float | list[float],
    current_history: str | None,
) -> Dataset:
    if isinstance(threshold, list):
        ds_list = []
        for th in threshold:
            config.threshold = th
            ds_list.append(_compute_ecad_index(index, config, current_history))
        result_ds = xr.concat(ds_list, dim="threshold")
    else:
        config.threshold = threshold
        result_ds = _compute_ecad_index(index, config, current_history)
    return result_ds


def _compute_user_index_dataset(
    config: IndexConfig, user_index: UserIndexDict
) -> Dataset:
    logging.info("Calculating user index.")
    result_ds = Dataset()
    deprecated_name = user_index.get("indice_name", None)
    if deprecated_name is not None:
        user_index["index_name"] = deprecated_name
        del user_index["indice_name"]
        log.deprecation_warning("indice_name", "index_name")
    user_indice_config = UserIndexConfig(
        **user_index,
        freq=config.freq,
        cf_vars=config._cf_variables,
        is_percent=config.is_percent,
        save_percentile=config.save_percentile,
    )
    user_indice_da = compute_user_index(user_indice_config)
    user_indice_da.attrs["units"] = _get_unit(config.out_unit, user_indice_da)
    if user_indice_config.calc_operation is CalcOperation.ANOMALY:
        # with anomaly time axis disappear
        result_ds[user_indice_config.index_name] = user_indice_da
        return result_ds
    user_indice_da, time_bounds = config.freq.post_processing(user_indice_da)
    result_ds[user_indice_config.index_name] = user_indice_da
    result_ds["time_bounds"] = time_bounds
    return result_ds


def _get_unit(output_unit: str | None, da: DataArray) -> str | None:
    da_unit = da.attrs.get("units", None)
    if da_unit is None:
        if output_unit is None:
            warn(
                "No unit computed or provided for the index was found. "
                "Use out_unit parameter to add one."
            )
            return ""
        else:
            return output_unit
    else:
        return da_unit


def _compute_ecad_index(
    index: EcadIndex, config: IndexConfig, former_history: str | None
) -> Dataset:
    logging.info(f"Calculating climate index: {index.short_name}")
    result_ds = Dataset()
    res = index.compute(config)
    if isinstance(res, tuple):
        da, per = res
    else:
        da, per = (res, None)
    da.attrs["units"] = _get_unit(config.out_unit, da)
    if config.threshold is not None:
        da.expand_dims({"threshold": config.threshold})
    if config.freq.post_processing is not None:
        resampled_da, time_bounds = config.freq.post_processing(da)
        result_ds[index.short_name] = resampled_da
        if time_bounds is not None:
            result_ds["time_bounds"] = time_bounds
            result_ds.time.attrs["bounds"] = "time_bounds"
    else:
        result_ds[index.short_name] = da
    if per is not None:
        per = per.squeeze("percentiles", drop=True).rename("percentiles")
        result_ds = xr.merge([result_ds, per])
    if former_history is None:
        former_history = da.attrs["history"]
    else:
        former_history = f"{former_history}\n{da.attrs['history']}"
    del da.attrs["history"]
    result_ds = _add_ecad_index_metadata(result_ds, config, index, former_history)
    return result_ds


def _add_ecad_index_metadata(
    result_ds: Dataset,
    config: IndexConfig,
    computed_index: EcadIndex,
    former_history: str,
) -> Dataset:
    result_ds.attrs.update(
        dict(
            title=_get_title(computed_index, config),
            references="ATBD of the ECA&D indices calculation"
            " (https://www.ecad.eu/documents/atbd.pdf)",
            institution="Climate impact portal (https://climate4impact.eu)",
            history=_get_history(config, former_history, computed_index, result_ds),
            source="",
            Conventions="CF-1.6",
        )
    )
    result_ds.lat.encoding["_FillValue"] = None
    result_ds.lon.encoding["_FillValue"] = None
    return result_ds


def _get_title(computed_index, config):
    if config.threshold is not None:
        return f"Index {computed_index.short_name} with user defined threshold"
    else:
        return f"ECA&D {computed_index.group.value} index {computed_index.short_name}"


def _get_history(config, former_history, indice_computed, result_ds):
    current_time = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    start_time = result_ds.time[0].dt.strftime("%m-%d-%Y").data[()]
    end_time = result_ds.time[-1].dt.strftime("%m-%d-%Y").data[()]
    return (
        f"{former_history}\n "
        f"{current_time} "
        f"Calculation of {indice_computed.short_name} "
        f"index({config.freq.description}) "
        f"from {start_time} to {end_time}."
    )


def _has_valid_unit(group: IndexGroup, da: DataArray) -> bool:
    if group == IndexGroup.SNOW:
        try:
            # todo: might be replaced by cf-xarray
            xclim.core.units.check_units.__wrapped__(da, "[length]")
        except xclim.core.utils.ValidationError:
            return False
    # For now we can delay to xclim other unit checks
    return True


def _guess_variable_names(
    in_var_name: str | list[str], index: EcadIndex | None, ds: Dataset
) -> list[str]:
    """
    Try to guess the variable names using the expected kind of variable for
    the index.
    """
    if isinstance(in_var_name, str):
        return [in_var_name]
    res = []
    index_variables = index.variables
    for indice_var in index_variables:
        for alias in indice_var:
            # check if dataset contains this alias
            if ds.get(alias, None) is not None and _has_valid_unit(
                index.group, ds[alias]
            ):
                res.append(alias)
                break
    if len(res) < len(index_variables):
        main_aliases = ", ".join(map(lambda v: v[0], index_variables))
        raise InvalidIcclimArgumentError(
            f"Index {index.short_name} needs the following variable(s)"
            f" [{main_aliases}], some of these were not recognized from the input."
            f" Use `var_name` parameter to explicitly use the data variable(s)"
            f" from your input dataset: {list(ds.data_vars)}."
        )
    return res



import numpy as np
import pandas as pd
import pytest
import xarray as xr

from icclim.models.cf_calendar import CfCalendar


class Test_CfCalendar:
    def test_error_lookup(self):
        with pytest.raises(TypeError):
            CfCalendar.lookup("NOPE!")

    @pytest.mark.parametrize(
        "cal",
        [
            "noleap",
            "360_day",
            "all_leap",
            "proleptic_gregorian",
            "julian",
            "standard",
            "none",
        ],
    )
    def test_success_lookup(self, cal):
        assert CfCalendar.lookup(cal).aliases[0] == cal

    def test_NO_LEAP(self):
        da = xr.DataArray(pd.date_range("2000", periods=100, freq="YS"), dims=["time"])
        res = CfCalendar.NO_LEAP.is_leap(da)
        np.testing.assert_array_equal(False, res)

    def test_DAYS_360(self):
        da = xr.DataArray(pd.date_range("2000", periods=100, freq="YS"), dims=["time"])
        res = CfCalendar.DAYS_360.is_leap(da)
        np.testing.assert_array_equal(False, res)

    def test_ALL_LEAP(self):
        da = xr.DataArray(pd.date_range("2000", periods=100, freq="YS"), dims=["time"])
        res = CfCalendar.ALL_LEAP.is_leap(da)
        np.testing.assert_array_equal(True, res)

    def test_PROLEPTIC_GREGORIAN(self):
        res_1 = CfCalendar.PROLEPTIC_GREGORIAN.is_leap(
            xr.DataArray(np.asarray([40, 1600]))
        )
        res_2 = CfCalendar.PROLEPTIC_GREGORIAN.is_leap(
            xr.DataArray(np.asarray([42, 1500, 1700]))
        )
        np.testing.assert_array_equal(True, res_1)
        np.testing.assert_array_equal(False, res_2)

    def test_JULIAN(self):
        res_1 = CfCalendar.JULIAN.is_leap(
            xr.DataArray(np.asarray([40, 1500, 1600, 1700]))
        )
        res_2 = CfCalendar.JULIAN.is_leap(xr.DataArray(np.asarray([42])))
        np.testing.assert_array_equal(True, res_1)
        np.testing.assert_array_equal(False, res_2)

    @pytest.mark.parametrize(
        "cal",
        [CfCalendar.STANDARD, CfCalendar.NONE],
    )
    def test_STANDARD(self, cal):
        res_1 = cal.is_leap(xr.DataArray(np.asarray([40, 1500, 1600])))
        res_2 = cal.is_leap(xr.DataArray(np.asarray([42, 1700])))
        np.testing.assert_array_equal(True, res_1)
        np.testing.assert_array_equal(False, res_2)



from __future__ import annotations

from enum import Enum
from typing import Callable, Optional, Tuple, Union

from xarray import DataArray

from icclim.clix_meta.clix_meta_indices import ClixMetaIndices
from icclim.ecad_functions import (
    cd,
    cdd,
    cfd,
    csdi,
    csu,
    cw,
    cwd,
    dtr,
    etr,
    fd,
    gd4,
    hd17,
    id,
    prcptot,
    r10mm,
    r20mm,
    r75p,
    r75ptot,
    r95p,
    r95ptot,
    r99p,
    r99ptot,
    rr1,
    rx1day,
    rx5day,
    sd,
    sd1,
    sd5cm,
    sd50cm,
    sdii,
    su,
    tg,
    tg10p,
    tg90p,
    tn,
    tn10p,
    tn90p,
    tnn,
    tnx,
    tr,
    tx,
    tx10p,
    tx90p,
    txn,
    txx,
    vdtr,
    wd,
    wsdi,
    ww,
)
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.constants import (
    ECAD_ATBD,
    MODIFIABLE_QUANTILE_WINDOW,
    MODIFIABLE_THRESHOLD,
    MODIFIABLE_UNIT,
    PR,
    QUANTILE_BASED,
    TAS,
    TAS_MAX,
    TAS_MIN,
)
from icclim.models.index_config import IndexConfig
from icclim.models.index_group import IndexGroup

ComputeIndexFun = Callable[
    [IndexConfig], Union[DataArray, Tuple[DataArray, Optional[DataArray]]]
]

clix_indices: ClixMetaIndices = ClixMetaIndices.get_instance()


class EcadIndex(Enum):
    """
    ECA&D indices.
        short_name: str
            The index name used in the output.
        compute: Callable
            The function to compute the index. It wraps Xclim functions.
        group: IndexGroup
            The index group category.
        variables: List[List[str]]
            The Cf variables needed to compute the index.
            The variable are individually described by a list of aliases.
        qualifiers: List[str]
            ``optional`` List of configuration to compute the index.
            Used internally to generate modules for C3S.
    """

    # Temperature
    TG = (
        "TG",
        lambda c: tg(c),
        IndexGroup.TEMPERATURE,
        [TAS],
    )
    TN = (
        "TN",
        lambda c: tn(c),
        IndexGroup.TEMPERATURE,
        [TAS_MIN],
    )
    TX = (
        "TX",
        lambda c: tx(c),
        IndexGroup.TEMPERATURE,
        [TAS_MAX],
    )
    DTR = (
        "DTR",
        lambda c: dtr(c),
        IndexGroup.TEMPERATURE,
        [TAS_MAX, TAS_MIN],
    )
    ETR = (
        "ETR",
        lambda c: etr(c),
        IndexGroup.TEMPERATURE,
        [TAS_MAX, TAS_MIN],
    )
    VDTR = (
        "vDTR",
        lambda c: vdtr(c),
        IndexGroup.TEMPERATURE,
        [TAS_MAX, TAS_MIN],
    )

    # Heat
    SU = (
        "SU",
        lambda c: su(c),
        IndexGroup.HEAT,
        [TAS_MAX],
        [MODIFIABLE_THRESHOLD],
    )
    TR = (
        "TR",
        lambda c: tr(c),
        IndexGroup.HEAT,
        [TAS_MIN],
        [MODIFIABLE_THRESHOLD],
    )
    WSDI = (
        "WSDI",
        lambda c: wsdi(c),
        IndexGroup.HEAT,
        [TAS_MAX],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_THRESHOLD],
    )
    TG90P = (
        "TG90p",
        lambda c: tg90p(c),
        IndexGroup.HEAT,
        [TAS],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TN90P = (
        "TN90p",
        lambda c: tn90p(c),
        IndexGroup.HEAT,
        [TAS_MIN],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TX90P = (
        "TX90p",
        lambda c: tx90p(c),
        IndexGroup.HEAT,
        [TAS_MAX],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TXX = (
        "TXx",
        lambda c: txx(c),
        IndexGroup.HEAT,
        [TAS_MAX],
    )
    TNX = (
        "TNx",
        lambda c: tnx(c),
        IndexGroup.HEAT,
        [TAS_MIN],
    )
    CSU = (
        "CSU",
        lambda c: csu(c),
        IndexGroup.HEAT,
        [TAS_MAX],
        [MODIFIABLE_THRESHOLD],
    )

    # Cold
    GD4 = (
        "GD4",
        lambda c: gd4(c),
        IndexGroup.COLD,
        [TAS],
        [MODIFIABLE_THRESHOLD],
    )
    FD = (
        "FD",
        lambda c: fd(c),
        IndexGroup.COLD,
        [TAS_MIN],
        [MODIFIABLE_THRESHOLD],
    )
    CFD = (
        "CFD",
        lambda c: cfd(c),
        IndexGroup.COLD,
        [TAS_MIN],
        [MODIFIABLE_THRESHOLD],
    )
    HD17 = (
        "HD17",
        lambda c: hd17(c),
        IndexGroup.COLD,
        [TAS],
        [MODIFIABLE_THRESHOLD],
    )
    ID = (
        "ID",
        lambda c: id(c),
        IndexGroup.COLD,
        [TAS_MAX],
        [MODIFIABLE_THRESHOLD],
    )
    TG10P = (
        "TG10p",
        lambda c: tg10p(c),
        IndexGroup.COLD,
        [TAS],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TN10P = (
        "TN10p",
        lambda c: tn10p(c),
        IndexGroup.COLD,
        [TAS_MIN],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TX10P = (
        "TX10p",
        lambda c: tx10p(c),
        IndexGroup.COLD,
        [TAS_MAX],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_UNIT],
    )
    TXN = (
        "TXn",
        lambda c: txn(c),
        IndexGroup.COLD,
        [TAS_MAX],
    )
    TNN = (
        "TNn",
        lambda c: tnn(c),
        IndexGroup.COLD,
        [TAS_MIN],
    )
    CSDI = (
        "CSDI",
        lambda c: csdi(c),
        IndexGroup.COLD,
        [TAS_MIN],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW, MODIFIABLE_THRESHOLD],
    )

    # Drought
    CDD = (
        "CDD",
        lambda c: cdd(c),
        IndexGroup.DROUGHT,
        [PR],
    )

    # Rain
    PRCPTOT = (
        "PRCPTOT",
        lambda c: prcptot(c),
        IndexGroup.RAIN,
        [PR],
    )
    RR1 = (
        "RR1",
        lambda c: rr1(c),
        IndexGroup.RAIN,
        [PR],
    )
    SDII = (
        "SDII",
        lambda c: sdii(c),
        IndexGroup.RAIN,
        [PR],
    )
    CWD = (
        "CWD",
        lambda c: cwd(c),
        IndexGroup.RAIN,
        [PR],
    )
    R10MM = (
        "R10mm",
        lambda c: r10mm(c),
        IndexGroup.RAIN,
        [PR],
    )
    R20MM = (
        "R20mm",
        lambda c: r20mm(c),
        IndexGroup.RAIN,
        [PR],
    )
    RX1DAY = (
        "RX1day",
        lambda c: rx1day(c),
        IndexGroup.RAIN,
        [PR],
    )
    RX5DAY = (
        "RX5day",
        lambda c: rx5day(c),
        IndexGroup.RAIN,
        [PR],
    )
    R75P = (
        "R75p",
        lambda c: r75p(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED, MODIFIABLE_UNIT],
    )
    R75PTOT = (
        "R75pTOT",
        lambda c: r75ptot(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED],
    )
    R95P = (
        "R95p",
        lambda c: r95p(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED, MODIFIABLE_UNIT],
    )
    R95PTOT = (
        "R95pTOT",
        lambda c: r95ptot(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED],
    )
    R99P = (
        "R99p",
        lambda c: r99p(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED, MODIFIABLE_UNIT],
    )
    R99PTOT = (
        "R99pTOT",
        lambda c: r99ptot(c),
        IndexGroup.RAIN,
        [PR],
        [QUANTILE_BASED],
    )

    # Snow
    SD = (
        "SD",
        lambda c: sd(c),
        IndexGroup.SNOW,
        [PR],
    )
    SD1 = (
        "SD1",
        lambda c: sd1(c),
        IndexGroup.SNOW,
        [PR],
    )
    SD5CM = (
        "SD5cm",
        lambda c: sd5cm(c),
        IndexGroup.SNOW,
        [PR],
    )
    SD50CM = (
        "SD50cm",
        lambda c: sd50cm(c),
        IndexGroup.SNOW,
        [PR],
    )

    # Compound (precipitation and temperature)
    CD = (
        "CD",
        lambda c: cd(c),
        IndexGroup.COMPOUND,
        [TAS, PR],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW],
    )
    CW = (
        "CW",
        lambda c: cw(c),
        IndexGroup.COMPOUND,
        [TAS, PR],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW],
    )
    WD = (
        "WD",
        lambda c: wd(c),
        IndexGroup.COMPOUND,
        [TAS, PR],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW],
    )
    WW = (
        "WW",
        lambda c: ww(c),
        IndexGroup.COMPOUND,
        [TAS, PR],
        [QUANTILE_BASED, MODIFIABLE_QUANTILE_WINDOW],
    )

    def __init__(
        self,
        short_name: str,
        compute: ComputeIndexFun,
        group: IndexGroup,
        variables: list[list[str]],
        qualifiers: list[str] = None,
    ):
        self.short_name = short_name
        self.compute = compute
        self.group = group
        self.variables = variables
        if qualifiers is None:
            qualifiers = []
        self.qualifiers = qualifiers
        definition = ""
        clix_index = clix_indices.lookup(short_name)
        if clix_index is not None:
            definition = clix_index["output"]["long_name"]
        self.definition = definition
        self.source = ECAD_ATBD

    @staticmethod
    def lookup(query: str) -> EcadIndex:
        if isinstance(query, EcadIndex):
            return query
        for e in EcadIndex:
            if e.short_name.upper() == query.upper():
                return e
        raise InvalidIcclimArgumentError(
            f"Unknown index {query}."
            f" You can list the available indices"
            f" with `icclim.list_indices()`."
        )

    @staticmethod
    def list() -> list[str]:
        """
        Get a a string list of ``EcadIndex`` enum's indices formatted in a readable
        fashion.
        """
        return [f"{i.group.value} | {i.short_name} | {i.definition}" for i in EcadIndex]



from __future__ import annotations

import xarray as xr
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.models.ecad_indices import EcadIndex


def read_dataset(
    data: str | list[str] | Dataset | DataArray,
    index: EcadIndex | None = None,
    var_names: str | list[str] | None = None,
) -> tuple[Dataset, bool, bool]:
    is_zarr = False
    if isinstance(data, Dataset):
        input_dataset = data
        chunk_da = False
    elif isinstance(data, DataArray):
        if index is None:
            # user index case
            if var_names is None or (
                isinstance(var_names, list) and len(var_names) > 1
            ):
                raise InvalidIcclimArgumentError(
                    "When the input is a DataArray, var_names must be a string."
                )
            if isinstance(var_names, list):
                var_names = var_names[0]
            data_name = var_names
        else:
            if len(index.variables) > 1:
                raise InvalidIcclimArgumentError(
                    f"Index {index.name} needs {len(index.variables)} variables."
                    f" Please provide them with an xarray.Dataset, a netCDF file or a"
                    f" zarr store."
                )
            data_name = index.variables[0][0]  # first alias of the unique variable
        input_dataset = data.to_dataset(name=data_name, promote_attrs=True)
        chunk_da = False
    elif isinstance(data, list):
        input_dataset = xr.open_mfdataset(data, parallel=True)
        chunk_da = True
    elif isinstance(data, str):
        if ".nc" in data:
            input_dataset = xr.open_dataset(data)
            chunk_da = True
        else:  # assume it's a zarr store
            input_dataset = xr.open_zarr(data)
            chunk_da = True
            is_zarr = True
    else:
        raise NotImplementedError("`in_files` format was not recognized.")
    return input_dataset, chunk_da, is_zarr


def update_to_standard_coords(ds: Dataset) -> tuple[Dataset, dict]:
    """
    Mutate input ds to use more icclim friendly coordinate name.
    """
    # TODO see if cf-xarray could replace this
    revert = {}
    if ds.coords.get("latitude") is not None:
        ds = ds.rename({"latitude": "lat"})
        revert.update({"lat": "latitude"})
    if ds.coords.get("longitude") is not None:
        ds = ds.rename({"longitude": "lon"})
        revert.update({"lon": "longitude"})
    if ds.coords.get("t") is not None:
        ds = ds.rename({"t": "time"})
        revert.update({"time": "t"})
    return ds, revert



from __future__ import annotations

import contextlib
import copy

import dask
import fsspec
import psutil
import xarray as xr
import zarr
from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from rechunker import rechunk
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset

import icclim.utils as utils
from icclim.icclim_exceptions import InvalidIcclimArgumentError
from icclim.icclim_logger import IcclimLogger
from icclim.pre_processing.input_parsing import read_dataset

TMP_STORE_1 = "icclim-tmp-store-1.zarr"
TMP_STORE_2 = "icclim-tmp-store-2.zarr"
DEFAULT_DASK_CONF = {
    "distributed.worker.memory.target": False,
    "distributed.worker.memory.spill": False,
    "distributed.worker.memory.pause": "0.95",
    "distributed.worker.memory.terminate": "0.98",
}

logger = IcclimLogger.get_instance()


def _get_mem_limit(factor: float = 0.9) -> int:
    # According to
    # https://github.com/pangeo-data/rechunker/issues/54#issuecomment-700748875
    # we should limit rechunk mem usage to around 0.9 and avoid spilling to disk
    if factor > 1 or factor < 0:
        raise ValueError(f"factor was {factor} but, it must be between 0 and 1.")
    try:
        import distributed

        max_sys_mem = (
            distributed.get_client()
            .submit(lambda: distributed.get_worker().memory_limit)
            .result()
        )
    except (ValueError, ImportError):
        # Assumes default scheduler is used
        max_sys_mem = psutil.virtual_memory().total
    return int(factor * max_sys_mem)


@contextlib.contextmanager
def create_optimized_zarr_store(
    in_files: str | list[str] | Dataset | DataArray,
    var_names: str | list[str],
    target_zarr_store_name: str = "icclim-target-store.zarr",
    keep_target_store: bool = False,
    chunking: dict[str, int] | None = None,
    filesystem: str | AbstractFileSystem = LocalFileSystem(),
) -> xr.Dataset:
    """
    -- EXPERIMENTAL FEATURE --
    API may changes without deprecation warning!

    Context manager to create an zarr store given an input netcdf or xarray structure.
    The execution may take a long time

    The result is rechunked according to `chunking` schema provided.
    By default, when leaving `chunking` to None, the resulting zarr store is NOT chunked
    on time dimension.
    This kind of chunking will significantly speed up the bootstrapping of
    percentiles for indices such as Tx90p, Tx10p, TN90p...
    But such chunking will most likely result in suboptimal performances for other
    indices.
    Actually, when computing indices where no bootstrap is needed,
    you should first try the computation without using `create_optimized_zarr_store`.
    If there are performance issues, you may want to use `create_optimized_zarr_store`
    with a dictionary of a better chunking schema than your current storage chunking.

    By default, `keep_target_store` being False, the resulting zarr store is destroyed
    when the context manager is exit.
    To keep the zarr store for futur uses set `keep_target_store` to True.

    The output is the resulting zarr store as a xarray Dataset.

    Examples
    --------

    >>> import icclim
    >>> with icclim.create_optimized_zarr_store(
    >>>                             in_files="tasmax.nc",
    >>>                             var_names="tasmax",
    >>>                             target_zarr_store_name="tasmax-store.zarr",
    >>>                             chunking={"time": 42, "lat": 42, "lon": 42},
    >>>                             ) as tasmax_opti:
    >>>      su_out = icclim.index(in_files = tasmax_opti, index_name = "su")

    Parameters
    ----------
    in_files : str | list[str] | Dataset | DataArray
        Absolute path(s) to NetCDF dataset(s), including OPeNDAP URLs,
        or path to zarr store, or xarray.Dataset or xarray.DataArray.
    var_names : str | list[str]
        List of data variable to include in the target zarr store.
        All other data variable are dropped.
        The coordinate variable are untouched and are part of the target zarr store.
    target_zarr_store_name : str
        Name of the target zarr store.
        Used to avoid overriding an existing zarr store.
    chunking : dict
        The target chunking schema.
    keep_target_store : bool
        Set to True to keep the target zarr store after the execution of the context
        manager.
        Set to False to remove the target zarr store once execution is finished.
        Default is False.
    filesystem :
        A fsspec filesystem where the zarr store will be created.

    Returns
    -------
    returns xr.Dataset opened on the newly created target zarr store.

    """
    try:
        if isinstance(filesystem, str):
            filesystem = fsspec.filesystem("file")
        _remove_stores(
            TMP_STORE_1, TMP_STORE_2, target_zarr_store_name, filesystem=filesystem
        )
        yield _unsafe_create_optimized_zarr_store(
            in_files,
            var_names,
            target_zarr_store_name,
            chunking,
            _get_mem_limit(),
            filesystem,
        )
    finally:
        stores_to_remove = [TMP_STORE_1, TMP_STORE_2]
        if not keep_target_store:
            stores_to_remove.append(target_zarr_store_name)
        _remove_stores(*stores_to_remove, filesystem=filesystem)


def _remove_stores(*stores, filesystem):
    for s in stores:
        try:
            filesystem.rm(s, recursive=True, maxdepth=100)
        except FileNotFoundError:
            pass


def _unsafe_create_optimized_zarr_store(
    in_files: str | list[str] | Dataset | DataArray,
    var_names: str | list[str],
    zarr_store_name: str,
    chunking: dict[str, int] | None,
    max_mem: int,
    filesystem: AbstractFileSystem,
):
    with dask.config.set(DEFAULT_DASK_CONF):
        logger.info("Rechunking in progress, this will take some time.")
        ds, _, is_zarr = read_dataset(in_files, index=None, var_names=var_names)
        # drop all non essential data variables
        ds = ds.drop_vars(filter(lambda v: v not in var_names, ds.data_vars.keys()))
        if len(ds.data_vars.keys()) == 0:
            raise InvalidIcclimArgumentError(
                f"The variable(s) {var_names} were not found in the dataset."
            )
        if _is_rechunking_unnecessary(ds, chunking):
            raise InvalidIcclimArgumentError(
                f"The given input is already chunked following {chunking}."
                f" It's unnecessary to rechunk data with"
                f" `create_optimized_zarr_store` here."
            )
        elif chunking is None:
            chunking = _build_default_chunking(ds)
        # It seems rechunker performs better when the dataset is first converted
        # to a zarr store, without rechunking anything.
        if not is_zarr:
            # needed to have unified chunk that can be written to zarr
            ds = ds.chunk("auto").unify_chunks()
            ds.to_zarr(TMP_STORE_1, mode="w")
            ds = xr.open_zarr(TMP_STORE_1)
        ds = ds.chunk(chunking)
        target_chunks = {}
        for data_var in ds.data_vars:
            ds[data_var].encoding = {}
            acc = {}
            for dim in ds[data_var].dims:
                acc.update({dim: utils._get_chunksizes(ds)[dim][0]})
            target_chunks.update({data_var: acc})
        for c in ds.coords:
            ds[c].encoding = {}
            target_chunks.update({c: None})
        rechunk(
            source=ds,
            target_chunks=target_chunks,
            max_mem=max_mem,
            target_store=zarr_store_name,
            temp_store=TMP_STORE_2,
        ).execute()
        _remove_stores(TMP_STORE_1, TMP_STORE_2, filesystem=filesystem)
        zarr.consolidate_metadata(zarr_store_name)
        return xr.open_zarr(zarr_store_name)


def _build_default_chunking(ds: Dataset) -> dict:
    # Leave dask find the best chunking schema for all dimensions but `dim`
    chunking = {d: "auto" for d in ds.dims}
    chunking["time"] = -1
    return chunking


def _is_rechunking_unnecessary(ds, chunking) -> bool:
    cp = copy.deepcopy(ds.chunks)
    if chunking is None:
        return len(ds.chunks["time"]) == 1
    else:
        return ds.chunk(chunking).chunks == cp



"""
All ECA&D functions. Each function wraps its xclim equivalent functions adding icclim
metadata to it.
"""
import re
from typing import Callable, Optional, Tuple
from warnings import warn

import numpy as np
import xarray as xr
import xclim.core.utils
from xarray import DataArray
from xclim import atmos, land
from xclim.core.calendar import percentile_doy, resample_doy
from xclim.core.units import convert_units_to

from icclim.models.cf_calendar import CfCalendar
from icclim.models.constants import IN_BASE_IDENTIFIER, PERCENTILES_COORD
from icclim.models.frequency import Frequency
from icclim.models.index_config import CfVariable, IndexConfig
from icclim.models.quantile_interpolation import QuantileInterpolation


def gd4(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tas.study_da,
        threshold=4.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.growing_degree_days,
    )


def cfd(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmin.study_da,
        threshold=0.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.consecutive_frost_days,
    )


def fd(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmin.study_da,
        threshold=0.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.frost_days,
    )


def hd17(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tas.study_da,
        threshold=17.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.heating_degree_days,
    )


def id(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmax.study_da,
        threshold=0.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.ice_days,
    )


def csdi(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    thresh = 10 if config.threshold is None else config.threshold
    return _compute_spell_duration(
        cf_var=config.tasmin,
        freq=config.freq.panda_freq,
        per_thresh=thresh,
        per_window=config.window,
        per_interpolation=config.interpolation,
        min_spell_duration=6,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.cold_spell_duration_index,
    )


def tg10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tas,
        freq=config.freq,
        tas_per_thresh=10,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tg10p,
    )


def tn10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tasmin,
        freq=config.freq,
        tas_per_thresh=10,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tn10p,
    )


def tx10p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tasmax,
        freq=config.freq,
        tas_per_thresh=10,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tx10p,
    )


def txn(config: IndexConfig) -> DataArray:
    result = atmos.tx_min(config.tasmax.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def tnn(config: IndexConfig) -> DataArray:
    result = atmos.tn_min(config.tasmin.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def cdd(config: IndexConfig) -> DataArray:
    result = atmos.maximum_consecutive_dry_days(
        config.pr.study_da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result


def su(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmax.study_da,
        threshold=25.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.tx_days_above,
    )


def tr(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmin.study_da,
        threshold=20.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.tropical_nights,
    )


def wsdi(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    thresh = 90 if config.threshold is None else config.threshold
    return _compute_spell_duration(
        cf_var=config.tasmax,
        freq=config.freq.panda_freq,
        per_thresh=thresh,
        per_window=config.window,
        per_interpolation=config.interpolation,
        min_spell_duration=6,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.warm_spell_duration_index,
    )


def tg90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tas,
        freq=config.freq,
        tas_per_thresh=90,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tg90p,
    )


def tn90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tasmin,
        freq=config.freq,
        tas_per_thresh=90,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tn90p,
    )


def tx90p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_temperature_percentile_index(
        cf_var=config.tasmax,
        freq=config.freq,
        tas_per_thresh=90,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
        callback=config.callback,
        xclim_index_fun=atmos.tx90p,
    )


def txx(config: IndexConfig) -> DataArray:
    result = atmos.tx_max(config.tasmax.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def tnx(config: IndexConfig) -> DataArray:
    result = atmos.tn_max(config.tasmin.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def csu(config: IndexConfig) -> DataArray:
    return _compute_threshold_index(
        da=config.tasmax.study_da,
        threshold=25.0 if config.threshold is None else config.threshold,
        freq=config.freq,
        xclim_index_fun=atmos.maximum_consecutive_warm_days,
    )


def prcptot(config: IndexConfig) -> DataArray:
    result = atmos.precip_accumulation(
        _filter_in_wet_days(config.pr.study_da, dry_day_value=0),
        freq=config.freq.panda_freq,
    )
    return result


def rr1(config: IndexConfig) -> DataArray:
    result = atmos.wetdays(
        config.pr.study_da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result


def sdii(config: IndexConfig) -> DataArray:
    result = atmos.daily_pr_intensity(
        config.pr.study_da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result


def cwd(config: IndexConfig) -> DataArray:
    result = atmos.maximum_consecutive_wet_days(
        config.pr.study_da, thresh="1.0 mm/day", freq=config.freq.panda_freq
    )
    return result


def r10mm(config: IndexConfig) -> DataArray:
    result = atmos.wetdays(
        config.pr.study_da, thresh="10 mm/day", freq=config.freq.panda_freq
    )
    return result


def r20mm(config: IndexConfig) -> DataArray:
    result = atmos.wetdays(
        config.pr.study_da, thresh="20 mm/day", freq=config.freq.panda_freq
    )
    return result


def rx1day(config: IndexConfig) -> DataArray:
    result = atmos.max_1day_precipitation_amount(
        config.pr.study_da, freq=config.freq.panda_freq
    )
    return result


def rx5day(config: IndexConfig) -> DataArray:
    result = atmos.max_n_day_precipitation_amount(
        config.pr.study_da, window=5, freq=config.freq.panda_freq
    )
    return result


def r75p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxp(
        pr=config.pr,
        freq=config.freq,
        pr_per_thresh=75.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
    )


def r75ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxptot(
        pr=config.pr,
        freq=config.freq.panda_freq,
        pr_per_thresh=75.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
    )


def r95p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxp(
        pr=config.pr,
        freq=config.freq,
        pr_per_thresh=95.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
    )


def r95ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxptot(
        pr=config.pr,
        freq=config.freq.panda_freq,
        pr_per_thresh=95.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
    )


def r99p(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxp(
        pr=config.pr,
        freq=config.freq,
        pr_per_thresh=99.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        is_percent=config.is_percent,
    )


def r99ptot(config: IndexConfig) -> Tuple[DataArray, Optional[DataArray]]:
    return _compute_rxxptot(
        pr=config.pr,
        freq=config.freq.panda_freq,
        pr_per_thresh=99.0,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
    )


def sd(config: IndexConfig) -> DataArray:
    """
    Climate index: Mean of daily Snow Depth (SD)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean of daily snow depth (cm)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.pr with the surface snow thickness
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = land.snow_depth(config.pr.study_da, freq=config.freq.panda_freq)
    return result


def sd1(config: IndexConfig) -> DataArray:
    """
    Climate index: Snow Days with at least 1 cm of snow (SD1)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Number of days with SD ≥ 1 cm (days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.pr with the surface snow thickness
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = land.snow_cover_duration(
        config.pr.study_da, thresh="1 cm", freq=config.freq.panda_freq
    )
    return result


def sd5cm(config: IndexConfig) -> DataArray:
    """
    Climate index: Snow Days with at least 5 cm of snow (SD5cm)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Number of days with SD ≥ 5 cm (days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.pr with the surface snow thickness
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = land.snow_cover_duration(
        config.pr.study_da, thresh="5 cm", freq=config.freq.panda_freq
    )
    return result


def sd50cm(config: IndexConfig) -> DataArray:
    """
    Climate index: Snow Days with at least 50 cm of snow (SD50cm)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Number of days with SD ≥ 50 cm (days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.pr with the surface snow thickness
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = land.snow_cover_duration(
        config.pr.study_da, thresh="50 cm", freq=config.freq.panda_freq
    )
    return result


def tg(config: IndexConfig) -> DataArray:
    """
    Climate index: mean of daily averaged(G) Temperature (TG)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean of daily averaged temperature (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tas
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.tg_mean(config.tas.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def tn(config: IndexConfig) -> DataArray:
    """
    Climate index: daily minimun(N) Temperature averaged (TN)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean of daily minimum temperature (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tasmin
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.tn_mean(config.tasmin.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def tx(config: IndexConfig) -> DataArray:
    """
    Climate index: daily maximum(X) Temperature averaged (TX)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean of daily maximum temperature (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tasmax
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.tx_mean(config.tasmax.study_da, freq=config.freq.panda_freq)
    result = convert_units_to(result, "°C")
    return result


def dtr(config: IndexConfig) -> DataArray:
    """
    Climate index: Diurnal Temperature Range (DTR)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean of diurnal temperature range (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tasmax
        config.cf_variables[1] with tasmin
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.daily_temperature_range(
        tasmax=config.tasmax.study_da,
        tasmin=config.tasmin.study_da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result


def etr(config: IndexConfig) -> DataArray:
    """
    Climate index: Extreme Temperature Range (ETR)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Intra-period extreme temperature range (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tasmax
        config.cf_variables[1] with tasmin
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.extreme_temperature_range(
        tasmax=config.tasmax.study_da,
        tasmin=config.tasmin.study_da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result


def vdtr(config: IndexConfig) -> DataArray:
    """
    Climate index:
        Mean absolute day-to-day difference in Diurnal temperature range (vDTR)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Mean absolute day-to-day difference in Diurnal temperature range (°C)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        config.cf_variables[0] with tasmax
        config.cf_variables[1] with tasmin
        config.freq
    Returns
    -------
    returns DataArray of the resulting index
    """
    result = atmos.daily_temperature_range_variability(
        tasmax=config.tasmax.study_da,
        tasmin=config.tasmin.study_da,
        freq=config.freq.panda_freq,
    )
    result.attrs["units"] = "°C"
    return result


def cd(config: IndexConfig) -> DataArray:
    """
    Climate index: Cold and Wet days (CD)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Days with TG < 25th percentile of daily mean temperature and RR < 25th percentile of
    daily precipitation sum (cold/dry days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        ``{cf_variables, window, interpolation, freq, save_percentile}``
    Returns
    -------
        If save_percentile is True,
            returns a Tuple of index_result, computed_percentiles
        Otherwise, returns the index_result
    """
    return compute_compound_index(
        tas=config.tas,
        pr=config.pr,
        freq=config.freq.panda_freq,
        tas_per_thresh=25,
        pr_per_thresh=25,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.cold_and_dry_days,
    )


def cw(config: IndexConfig) -> DataArray:
    """
    Climate index: Cold and Wet days (CD)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Days with TG < 25th percentile of daily mean temperature and RR > 75th percentile of
    daily precipitation sum (cold/wet days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        ``{cf_variables, window, interpolation, freq, save_percentile}``
    Returns
    -------
        If save_percentile is True,
            returns a Tuple of index_result, computed_percentiles
        Otherwise, returns the index_result
    """
    return compute_compound_index(
        tas=config.tas,
        pr=config.pr,
        freq=config.freq.panda_freq,
        tas_per_thresh=25,
        pr_per_thresh=75,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.cold_and_wet_days,
    )


def wd(config: IndexConfig) -> DataArray:
    """
    Climate index: Warm and Dry days (WD)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Days with TG > 75th percentile of daily mean temperature and RR <25th percentile
    of daily precipitation sum (warm/dry days)

    Parameters
    ----------
    config : IndexConfig
        The configuration necessary to compute this index.
        For this index the following fields must be filled:
        ``{cf_variables, window, interpolation, freq, save_percentile}``
    Returns
    -------
        If save_percentile is True,
            returns a Tuple made of  index_result, computed_percentiles
        Otherwise, returns the index_result
    """
    return compute_compound_index(
        tas=config.tas,
        pr=config.pr,
        freq=config.freq.panda_freq,
        tas_per_thresh=75,
        pr_per_thresh=25,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.warm_and_dry_days,
    )


def ww(config: IndexConfig) -> DataArray:
    """
    Climate index: Warm and Wet days (WW)

    Source: ECA&D, Algorithm Theoretical Basis Document (ATBD) v11

    Days with TG > 75th percentile of daily mean temperature and RR > 75th percentile of
    daily precipitation sum (warm/wet days)
    """
    return compute_compound_index(
        tas=config.tas,
        pr=config.pr,
        freq=config.freq.panda_freq,
        tas_per_thresh=75,
        pr_per_thresh=75,
        per_window=config.window,
        per_interpolation=config.interpolation,
        save_percentile=config.save_percentile,
        callback=config.callback,
        xclim_index_fun=atmos.warm_and_wet_days,
    )


def _can_run_bootstrap(cf_var: CfVariable) -> bool:
    """
    Avoid bootstrapping if there is one single year overlapping or no year overlapping
    or all year overlapping.
    """
    study_years = np.unique(cf_var.study_da.indexes.get("time").year)
    overlapping_years = np.unique(
        cf_var.study_da.sel(time=_get_ref_period_slice(cf_var.reference_da))
        .indexes.get("time")
        .year
    )
    return len(overlapping_years) > 1 and len(overlapping_years) < len(study_years)


def _get_ref_period_slice(da: DataArray) -> slice:
    time_length = len(da.time)
    return slice(*(da.time[0 :: time_length - 1].dt.strftime("%Y-%m-%d").values))


def _to_percent(da: DataArray, sampling_freq: Frequency) -> DataArray:
    if sampling_freq == Frequency.MONTH:
        da = da / da.time.dt.daysinmonth * 100
    elif sampling_freq == Frequency.YEAR:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 366
        coef[~leap_years] = 365
        da = da / coef
    elif sampling_freq == Frequency.AMJJAS:
        da = da / 183
    elif sampling_freq == Frequency.ONDJFM:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 183
        coef[~leap_years] = 182
        da = da / coef
    elif sampling_freq == Frequency.DJF:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[leap_years] = 91
        coef[~leap_years] = 90
        da = da / coef
    elif sampling_freq in [Frequency.MAM, Frequency.JJA]:
        da = da / 92
    elif sampling_freq == Frequency.SON:
        da = da / 91
    else:
        # TODO improve this for custom resampling
        warn(
            "For now, '%' unit can only be used when `slice_mode` is one of: "
            "{MONTH, YEAR, AMJJAS, ONDJFM, DJF, MAM, JJA, SON}."
        )
        return da
    da.attrs["units"] = "1"
    return da


def _is_leap_year(da: DataArray) -> np.ndarray:
    time_index = da.indexes.get("time")
    if isinstance(time_index, xr.CFTimeIndex):
        cf_calendar: CfCalendar = CfCalendar.lookup(time_index.calendar)
        return cf_calendar.is_leap(da.time.dt.year)
    else:
        return da.time.dt.is_leap_year


def _add_bootstrap_meta(result: DataArray, per: DataArray) -> DataArray:
    result.attrs[IN_BASE_IDENTIFIER] = per.climatology_bounds
    return result


def _compute_percentile_doy(
    da: DataArray,
    percentile: float,
    window: int = 5,
    interpolation=QuantileInterpolation.MEDIAN_UNBIASED,
    callback: Callable = None,
) -> DataArray:
    per = percentile_doy(
        da,
        window,
        percentile,
        alpha=interpolation.alpha,
        beta=interpolation.beta,
    )
    if callback is not None:
        callback(50)
    return per


def _compute_percentile_over_period(
    arr: DataArray, interpolation: QuantileInterpolation, percentiles: float
) -> DataArray:
    return xr.apply_ufunc(
        xclim.core.utils.calc_perc,
        arr,
        input_core_dims=[["time"]],
        output_core_dims=[[PERCENTILES_COORD]],
        kwargs=dict(
            percentiles=[percentiles],
            alpha=interpolation.alpha,
            beta=interpolation.beta,
        ),
        dask="parallelized",
        output_dtypes=[arr.dtype],
        dask_gufunc_kwargs=dict(output_sizes={PERCENTILES_COORD: len([percentiles])}),
    )


def _filter_in_wet_days(da: DataArray, dry_day_value: float):
    """
    Turns non wet days to NaN.
    dry_day_value may be Nan or 0.
    """
    precip = convert_units_to(da, "mm/d")
    return precip.where(precip > 1, dry_day_value)


def _compute_threshold_index(
    da: DataArray,
    threshold: float,
    freq: Frequency,
    xclim_index_fun: Callable,
) -> DataArray:
    result = xclim_index_fun(da, thresh=f"{threshold} °C", freq=freq.panda_freq)
    return result


def _compute_spell_duration(
    cf_var: CfVariable,
    freq: str,
    per_window: int,
    per_thresh: float,
    per_interpolation: QuantileInterpolation,
    min_spell_duration: int,
    save_percentile: bool,
    callback: Callable,
    xclim_index_fun: Callable,
) -> Tuple[DataArray, Optional[DataArray]]:
    per = _compute_percentile_doy(
        cf_var.reference_da,
        per_thresh,
        per_window,
        per_interpolation,
        callback,
    )
    run_bootstrap = _can_run_bootstrap(cf_var)
    result = xclim_index_fun(
        cf_var.study_da,
        per,
        window=min_spell_duration,
        freq=freq,
        bootstrap=run_bootstrap,
    )
    result = result.squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if save_percentile:
        return result, per
    result.attrs["description"] = re.sub(
        r"\s\w+th\spercentile",
        f" {per_thresh}th percentile",
        result.attrs.get("description"),
    )
    return result, None


def compute_compound_index(
    tas: CfVariable,
    pr: CfVariable,
    freq: str,
    tas_per_thresh: int,
    pr_per_thresh: int,
    per_window: int,
    per_interpolation: QuantileInterpolation,
    save_percentile: bool,
    callback: Callable,
    xclim_index_fun: Callable,
) -> DataArray:
    """CD, CW, WD, WW

    Parameters
    ----------
    tas : CfVariable
        CfVariable of tas variable.
    in_base_tas : DataArray
        DataArray of tas in_base variable.
        Used for percentile calculation only.
        Can overlap with tas.
    pr : DataArray
        DataArray of pr variable.
    in_base_pr : DataArray
        DataArray of pr in_base variable.
        Used for percentile calculation only.
        Can overlap with pr.
    window : int
        window in days used to computed each percentile.
        default is 5.
    freq : str
        pandas like frequency.
        Used to determine the time slices of the output.
        Default is "YS" as in Year Start.
    save_percentile : bool
        Flag to include coordinate variable including the computed percentiles.
        Does not contain the bootstrapped percentiles.
        Default is False.
    interpolation :QuantileInterpolation
        percentile interpolation method, default is MEDIAN_UNBIASED, a.k.a method 8.
    callback : Callable
        callback receiving an integer, may serve as a loading bar.

    Returns
    -------
        If save_percentile is True,
            returns a Tuple of index_result, computed_percentiles
        Otherwise, returns the index_result
    """
    tas_per = _compute_percentile_doy(
        tas.reference_da,
        tas_per_thresh,
        per_window,
        per_interpolation,
        callback,
    )
    tas_per = tas_per.squeeze(PERCENTILES_COORD, drop=True)
    pr_in_base = _filter_in_wet_days(pr.reference_da, dry_day_value=np.NAN)
    pr_out_of_base = _filter_in_wet_days(pr.study_da, dry_day_value=0)
    pr_per = _compute_percentile_doy(
        pr_in_base,
        pr_per_thresh,
        per_window,
        per_interpolation,
        callback,
    )
    pr_per = pr_per.squeeze(PERCENTILES_COORD, drop=True)
    result = xclim_index_fun(tas.study_da, tas_per, pr_out_of_base, pr_per, freq=freq)
    if save_percentile:
        # FIXME, not consistent with other percentile based indices
        #        We should probably return a Tuple (res, [tas_per, pr_per])
        #        However, here the percentiles use the existing time dimension and not
        #        doy
        result.coords["tas_per"] = resample_doy(tas_per, result)
        result.coords["pr_per"] = resample_doy(pr_per, result)
    return result


def _compute_rxxptot(
    pr: CfVariable,
    freq: str,
    pr_per_thresh: float,
    per_interpolation: QuantileInterpolation,
    save_percentile: bool,
) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(pr.reference_da, dry_day_value=np.nan)
    per = _compute_percentile_over_period(
        base_wet_days, per_interpolation, pr_per_thresh
    )
    result = atmos.fraction_over_precip_thresh(
        pr.study_da,
        per,
        thresh="1 mm/day",
        freq=freq,
        bootstrap=False,
    ).squeeze(PERCENTILES_COORD, drop=True)
    result = result * 100
    result.attrs["units"] = "%"
    if save_percentile:
        return result, per
    return result, None


def _compute_rxxp(
    pr: CfVariable,
    freq: Frequency,
    pr_per_thresh: float,
    per_interpolation: QuantileInterpolation,
    save_percentile: bool,
    is_percent: bool,
) -> Tuple[DataArray, Optional[DataArray]]:
    base_wet_days = _filter_in_wet_days(pr.reference_da, dry_day_value=np.nan)
    per = _compute_percentile_over_period(
        base_wet_days, per_interpolation, pr_per_thresh
    )
    result = atmos.days_over_precip_thresh(
        pr.study_da,
        per,
        thresh="1 mm/day",
        freq=freq.panda_freq,
        bootstrap=False,
    )
    result = result.squeeze(PERCENTILES_COORD, drop=True)
    if is_percent:
        result = _to_percent(result, freq)
    if save_percentile:
        return result, per
    return result, None


def _compute_temperature_percentile_index(
    cf_var: CfVariable,
    freq: Frequency,
    tas_per_thresh: int,
    per_window: int,
    per_interpolation: QuantileInterpolation,
    save_percentile: bool,
    is_percent: bool,
    callback: Callable,
    xclim_index_fun: Callable,
) -> Tuple[DataArray, Optional[DataArray]]:
    run_bootstrap = _can_run_bootstrap(cf_var)
    per = _compute_percentile_doy(
        cf_var.reference_da,
        tas_per_thresh,
        per_window,
        per_interpolation,
        callback,
    ).compute()
    result = xclim_index_fun(
        cf_var.study_da,
        per,
        freq=freq.panda_freq,
        bootstrap=run_bootstrap,
    ).squeeze(PERCENTILES_COORD, drop=True)
    if run_bootstrap:
        result = _add_bootstrap_meta(result, per)
    if is_percent:
        result = _to_percent(result, freq)
    if save_percentile:
        return result, per
    return result, None
